/*!
 *  @file EEPROM_settings.cpp
 *
 *
 * 	EEPROM using buffer with fault tolerant backups
 *
 *
 */

#include <EEPROM.h>
#include <FastCRC.h>
#include <FastCRC_cpu.h>
#include <FastCRC_tables.h>
#include <Arduino.h>

#include "EEPROM_settings.h"
#include "St_ism303dac.h"

SettingsBuffer_t SettingsBuffer;

EEPROM_settings::EEPROM_settings(void) 
{

}

void EEPROM_settings::Default_EEPROM_Settings (int mode)
{
    if (mode & 1)
    {
        SettingsBuffer.ISM303_INT_acc_x_minus100pcReading = -ACC_100pc;
        SettingsBuffer.ISM303_INT_acc_x_plus100pcReading  = ACC_100pc;
        SettingsBuffer.ISM303_INT_acc_y_minus100pcReading = -ACC_100pc;
        SettingsBuffer.ISM303_INT_acc_y_plus100pcReading  = ACC_100pc;
        SettingsBuffer.ISM303_INT_acc_z_minus100pcReading = -ACC_100pc;
        SettingsBuffer.ISM303_INT_acc_z_plus100pcReading  = ACC_100pc;
    
        SettingsBuffer.ISM303_INT_mag_x_minus100pcReading = -MAG_100pc;
        SettingsBuffer.ISM303_INT_mag_x_plus100pcReading  = MAG_100pc;
        SettingsBuffer.ISM303_INT_mag_y_minus100pcReading = -MAG_100pc;
        SettingsBuffer.ISM303_INT_mag_y_plus100pcReading  = MAG_100pc;
        SettingsBuffer.ISM303_INT_mag_z_minus100pcReading = -MAG_100pc;
        SettingsBuffer.ISM303_INT_mag_z_plus100pcReading  = MAG_100pc;
    }
    if (mode & 2)
    {
        for (int i=0; i<N_MagneticDensitySensors; i++)
        {
            SettingsBuffer.MagneticSensors0pcReadings[i] = 0;
            SettingsBuffer.MagneticSensors100pcReadings[i] = 16384;
        }
    }
    if (mode & 4)
    {
        SettingsBuffer.ControlMode = 0;;
        SettingsBuffer.RPMsetFast = 16384 * 70 / 100;
        SettingsBuffer.RPMsetNormal = 16384 * 10 / 100;;
        SettingsBuffer.TautoTrigger = 0;
        SettingsBuffer.TFoffset = 0;
        SettingsBuffer.ParkOrientation = 0;
        SettingsBuffer.Vmin = 20;
        SettingsBuffer.Vmax = 4000;
        SettingsBuffer.d_MB_MC = 1370;
        SettingsBuffer.TautoTurn = 300;
        SettingsBuffer.BitFlags_0 = 0;
     
        SettingsBuffer.d_MD_ME = 5000;
        SettingsBuffer.V_ratio = 456789;
        SettingsBuffer.spare1 = 0;
        SettingsBuffer.spare2 = 0;
        SettingsBuffer.spare3 = 0;
        SettingsBuffer.spare4 = 0;
    }
    Write_EEPROM_Settings ();
}

void EEPROM_settings::Write_EEPROM_Settings ()
{
    AddCRC32 ();
    for (int b=0; b<Nbackups; b++)
        Save_EEPROM_Settings (b);
}

bool EEPROM_settings::Read_EEPROM_Settings ()
{
    int b=0;
    while (b < Nbackups)
    {
        if (Load_EEPROM_Settings(b))
            return true;
        b++;
    }

    Default_EEPROM_Settings (0xffff);
    return false;
}

bool EEPROM_settings::Load_EEPROM_Settings (int WhichBuffer)
{
    int a = EEPROM_Settings_start_address + WhichBuffer * SettingsBufferSize;
    uint8_t *p = (uint8_t*)&SettingsBuffer;
    
    for (uint32_t i=0; i<SettingsBufferSize; i++)
        *p++ = EEPROM.read (a++);
    return CheckCRC32 ();   
}

void EEPROM_settings::Save_EEPROM_Settings (int WhichBuffer)
{
    AddCRC32 ();
    
    int a = EEPROM_Settings_start_address + WhichBuffer * SettingsBufferSize;
    uint8_t *p = (uint8_t*)&SettingsBuffer;
    
    for (uint32_t i=0; i<SettingsBufferSize; i++)
        EEPROM.write (a++, *p++);
}

void EEPROM_settings::AddCRC32 ()
{
    FastCRC32 crc;
    SettingsBuffer.crc32value = crc.crc32((uint8_t*)&SettingsBuffer, SettingsBufferSize-4);
}

bool EEPROM_settings::CheckCRC32 ()
{
    FastCRC32 crc;
    uint32_t crc_expected = crc.crc32((uint8_t*)&SettingsBuffer, SettingsBufferSize-4);
    return (crc_expected == SettingsBuffer.crc32value);
}

void EEPROM_settings::show_EEPROM_settings_bytes (int WhichBuffer)
{
    int a = EEPROM_Settings_start_address + WhichBuffer * SettingsBufferSize;
 
    Serial.printf ("buffer %d from 0x%03x to 0x%03x:\n", WhichBuffer, a, a+SettingsBufferSize-1);
    for (uint32_t i=0; i<SettingsBufferSize; i++)
    {
        Serial.printf ("%02x ", EEPROM.read (a++));
        if (((i % 16)==15) || (i == (SettingsBufferSize-1)))
            Serial.printf ("\n");
    }
    
}

void EEPROM_settings::show_all_EEPROM_settings_bytes ()
{
    for (int b=0; b<Nbackups; b++)
        show_EEPROM_settings_bytes (b);
}
