/*!
 *  @file INA260.cpp
 *
 *
 * 	I2C Driver for the INA260 I2C Current and Power sensor
 *
 *
 *  @section dependencies Dependencies
 *  Based on what once was the adafruit INA260 libraty - but we went Teensy4 with DMA, FIFOs etc...
 *
 *  @section author Author
 *
 *  Bryan Siepert for Adafruit Industries
 *
 * 	@section license License
 *
 * 	BSD (see license.txt)
 *
 */

#include "Arduino.h"

#include "INA260.h"

/*!
 *    @brief  Instantiates a new INA260 class
 */
INA260::INA260(void) {}

/*!
 *    @brief  Sets up the HW
 *    @param  i2c_address
 *            The I2C address to be used.
 *    @param  theWire
 *            The Wire object to be used for I2C connections.
 *    @return True if initialization was successful, otherwise false.
 */
bool INA260::begin(IMX_RT1060_I2CMaster& master, uint8_t i2c_slave_address) 
{
    i2c_dev = new I2CDevice(master, i2c_slave_address, _BIG_ENDIAN);

    if (i2c_dev->read(INA260_REG_MFG_UID, &ManuafacturerID_Register, false, true)) 
    {
        if (ManuafacturerID_Register != expected_manufacturer_id) 
        {
            Serial.printf("ERROR: Manufacturer ID is 0x%X. Expected 0x%X.\n", ManuafacturerID_Register, expected_manufacturer_id);
            return false;
        }
    } 
    else 
    {
        Serial.printf("ERROR: Failed to read manufacturer id register value");
        return false;
    }

    if (i2c_dev->read(INA260_REG_DIE_UID, &DieID_Register, false, true)) 
    {
        if (DieID_Register != expected_die_id) 
        {
            Serial.printf("ERROR: Die ID is 0x%X. Expected 0x%X.\n", DieID_Register, expected_die_id);
            return false;
        }
    } 
    else 
    {
        Serial.printf("ERROR: Failed to read manufacturer id register value");
        return false;
    }

    reset_device();
    delay(2); // delay 2ms to give time for first measurement to finish
    upload_all_registers ();

    isOK = true;
    return true;
}

void INA260::upload_all_registers (void)
{
    i2c_dev->read(INA260_REG_CONFIG, &ConfigurationRegister, false, true);
    i2c_dev->read(INA260_REG_CURRENT, &CurrentRegister, false, true);
    i2c_dev->read(INA260_REG_BUSVOLTAGE, &BusVoltageRegister, false, true);
    i2c_dev->read(INA260_REG_POWER, &PowerRegister, false, true);
    i2c_dev->read(INA260_REG_MASK_ENABLE, &Mask_EnableRegister, false, true);
    i2c_dev->read(INA260_REG_ALERT_LIMIT, &AlertLimitRegister, true, true);
}

void INA260::show_registers (String st)
{
    Serial.println(st);
    Serial.printf("INA260_REG_CONFIG is 0x%X\n", ConfigurationRegister);
    Serial.printf("INA260_REG_CURRENT is 0x%X\n", CurrentRegister);
    Serial.printf("INA260_REG_BUSVOLTAGE is 0x%X\n", BusVoltageRegister);
    Serial.printf("INA260_REG_POWER is 0x%X\n", PowerRegister);
    Serial.printf("INA260_REG_MASK_ENABLE is 0x%X\n", Mask_EnableRegister);
    Serial.printf("INA260_REG_ALERT_LIMIT is 0x%X\n", AlertLimitRegister);
    Serial.println(st);
}

/**************************************************************************/
/*!
    @brief Resets the harware. All registers are set to default values,
    the same as a power-on reset.
*/
/**************************************************************************/
void INA260::reset_device(void) 
{
    ConfigurationRegister |= (1 << 15);
    i2c_dev->write(INA260_REG_CONFIG, ConfigurationRegister, true, true);
    ConfigurationRegister &= ~(1 << 15);
}
/**************************************************************************/
/*!
    @brief Reads and scales the current value of the Current register.
    @return The current current measurement in mA
*/
/**************************************************************************/
float INA260::readCurrent(void) 
{
    i2c_dev->read(INA260_REG_CURRENT, &CurrentRegister, true, true);
    return (int16_t)CurrentRegister * 1.25;
}
/**************************************************************************/
/*!
    @brief Reads and scales the current value of the Bus Voltage register.
    @return The current bus voltage measurement in mV
*/
/**************************************************************************/
float INA260::readBusVoltage(void) 
{
    i2c_dev->read(INA260_REG_BUSVOLTAGE, &BusVoltageRegister, true, true);
    return BusVoltageRegister * 1.25;
}
/**************************************************************************/
/*!
    @brief Reads and scales the current value of the Power register.
    @return The current Power calculation in mW
*/
/**************************************************************************/
float INA260::readPower(void) 
{
    i2c_dev->read(INA260_REG_POWER, &PowerRegister, true, true);
    return PowerRegister * 10;
}

void INA260::readCurrVoltPower (float *current, float *voltage, float *power)
{
    i2c_dev->read(INA260_REG_CURRENT, &CurrentRegister, false, true);
    i2c_dev->read(INA260_REG_BUSVOLTAGE, &BusVoltageRegister, false, true);
    i2c_dev->read(INA260_REG_POWER, &PowerRegister, true, true);
    *current = (int16_t)CurrentRegister * 1.25;
    *voltage = BusVoltageRegister * 1.25;
    *power = PowerRegister * 10;
}

void INA260::readCurrVoltPower (float *current, float *voltage, float *power, INA260_MeasurementMode new_mode)
{
    // values from previous readout should be ready now
    *current = (int16_t)CurrentRegister * 1.25;
    *voltage = BusVoltageRegister * 1.25;
    *power = PowerRegister * 10;

    // trigger the next readout cycle, but do not wait until we get its result (I2C bus need its time, let the CPU do other useful stuff meanwhile)
    i2c_dev->read(INA260_REG_CURRENT, &CurrentRegister, true, false);
    i2c_dev->read(INA260_REG_BUSVOLTAGE, &BusVoltageRegister, true, false);
    i2c_dev->read(INA260_REG_POWER, &PowerRegister, true, false);
    
    ConfigurationRegister &= ~7;
    ConfigurationRegister |= new_mode;
    i2c_dev->write(INA260_REG_CONFIG, ConfigurationRegister, true, false);
}

/**************************************************************************/
/*!
    @brief Returns the current measurement mode
    @return The current mode
*/
/**************************************************************************/
INA260_MeasurementMode INA260::getMode(void) 
{
    i2c_dev->read(INA260_REG_CONFIG, &ConfigurationRegister, true, true);
    return (INA260_MeasurementMode)(ConfigurationRegister & 7);
}
/**************************************************************************/
/*!
    @brief Sets a new measurement mode
    @param new_mode
           The new mode to be set
*/
/**************************************************************************/
void INA260::setMode(INA260_MeasurementMode new_mode) 
{
    ConfigurationRegister &= ~7;
    ConfigurationRegister |= new_mode;
    i2c_dev->write(INA260_REG_CONFIG, ConfigurationRegister, true, true);
}
/**************************************************************************/
/*!
    @brief Reads the current number of averaging samples
    @return The current number of averaging samples
*/
/**************************************************************************/
INA260_AveragingCount INA260::getAveragingCount(void) 
{
    i2c_dev->read(INA260_REG_CONFIG, &ConfigurationRegister, true, true);
    return (INA260_AveragingCount)((ConfigurationRegister >> 9) & 7);
}
/**************************************************************************/
/*!
    @brief Sets the number of averaging samples
    @param count
           The number of samples to be averaged
*/
/**************************************************************************/
void INA260::setAveragingCount(INA260_AveragingCount count) 
{
    ConfigurationRegister &= ~(7 << 9);
    ConfigurationRegister |= ((uint16_t)count << 9);
    i2c_dev->write(INA260_REG_CONFIG, ConfigurationRegister, true, true);
}
/**************************************************************************/
/*!
    @brief Reads the current current conversion time
    @return The current current conversion time
*/
/**************************************************************************/
INA260_ConversionTime INA260::getCurrentConversionTime(void) 
{
    i2c_dev->read(INA260_REG_CONFIG, &ConfigurationRegister, true, true);
    return (INA260_ConversionTime)((ConfigurationRegister >> 3) & 7);
}
/**************************************************************************/
/*!
    @brief Sets the current conversion time
    @param time
           The new current conversion time
*/
/**************************************************************************/
void INA260::setCurrentConversionTime(INA260_ConversionTime conv_time) 
{
    ConfigurationRegister &= ~(7 << 3);
    ConfigurationRegister |= ((uint16_t)conv_time << 3);
    i2c_dev->write(INA260_REG_CONFIG, ConfigurationRegister, true, true);
}
/**************************************************************************/
/*!
    @brief Reads the current bus voltage conversion time
    @return The current bus voltage conversion time
*/
/**************************************************************************/
INA260_ConversionTime INA260::getVoltageConversionTime(void) 
{
    i2c_dev->read(INA260_REG_CONFIG, &ConfigurationRegister, true, true);
    return (INA260_ConversionTime)((ConfigurationRegister >> 6) & 7);
}
/**************************************************************************/
/*!
    @brief Sets the bus voltage conversion time
    @param time
           The new bus voltage conversion time
*/
/**************************************************************************/
void INA260::setVoltageConversionTime(INA260_ConversionTime conv_time) 
{
    ConfigurationRegister &= ~(7 << 6);
    ConfigurationRegister |= ((uint16_t)conv_time << 6);
    i2c_dev->write(INA260_REG_CONFIG, ConfigurationRegister, true, true);
}

/**************************************************************************/
/*!
    @brief Checks if the most recent one shot measurement has completed
    @return true if the conversion has completed
*/
/**************************************************************************/
bool INA260::conversionReady(void) 
{
    i2c_dev->read(INA260_REG_MASK_ENABLE, &Mask_EnableRegister, true, true);
    return (Mask_EnableRegister >> 3) & 1;
}
/**************************************************************************/
/*!
    @brief Reads the current parameter that asserts the ALERT pin
    @return The current parameter that asserts the ALERT PIN
*/
/**************************************************************************/
INA260_AlertType INA260::getAlertType(void) 
{
    i2c_dev->read(INA260_REG_MASK_ENABLE, &Mask_EnableRegister, true, true);
    return (INA260_AlertType)((Mask_EnableRegister >> 10) & 0x3f);
}
/**************************************************************************/
/*!
    @brief Sets which parameter asserts the ALERT pin
    @param alert
           The parameter which asserts the ALERT pin
*/
/**************************************************************************/
void INA260::setAlertType(INA260_AlertType alert) 
{
    Mask_EnableRegister &= ~(0x3f << 10);
    Mask_EnableRegister |= ((uint16_t)alert << 10);
    i2c_dev->write(INA260_REG_MASK_ENABLE, Mask_EnableRegister, true, true);
}
/**************************************************************************/
/*!
    @brief Reads the current alert limit setting
    @return The current bus alert limit setting
*/
/**************************************************************************/
float INA260::getAlertLimit(void) 
{
    i2c_dev->read(INA260_REG_ALERT_LIMIT, &AlertLimitRegister, true, true);
    return (float)AlertLimitRegister * 1.25;
}
/**************************************************************************/
/*!
    @brief Sets the Alert Limit
    @param limit
           The new limit that triggers the alert
*/
/**************************************************************************/
void INA260::setAlertLimit(float limit) 
{
    AlertLimitRegister = (int16_t)(limit / 1.25);
    i2c_dev->write(INA260_REG_ALERT_LIMIT, AlertLimitRegister, true, true);
}
/**************************************************************************/
/*!
    @brief Reads the current alert polarity setting
    @return The current bus alert polarity setting
*/
/**************************************************************************/
INA260_AlertPolarity INA260::getAlertPolarity(void) 
{
    i2c_dev->read(INA260_REG_MASK_ENABLE, &Mask_EnableRegister, true, true);
    return (INA260_AlertPolarity)((Mask_EnableRegister >> 1) & 1);
}
/**************************************************************************/
/*!
    @brief Sets Alert Polarity Bit
    @param polarity
           The polarity of the alert pin
*/
/**************************************************************************/
void INA260::setAlertPolarity(INA260_AlertPolarity polarity) 
{
    Mask_EnableRegister &= ~(1 << 1);
    Mask_EnableRegister |= ((uint16_t)polarity << 1);
    i2c_dev->write(INA260_REG_MASK_ENABLE, Mask_EnableRegister, true, true);
}
/**************************************************************************/
/*!
    @brief Reads the current alert latch setting
    @return The current bus alert latch setting
*/
/**************************************************************************/
INA260_AlertLatch INA260::getAlertLatch(void) 
{
    i2c_dev->read(INA260_REG_MASK_ENABLE, &Mask_EnableRegister, true, true);
    return (INA260_AlertLatch)((Mask_EnableRegister >> 0) & 1);
}
/**************************************************************************/
/*!
    @brief Sets Alert Latch Bit
    @param state
           The parameter which asserts the ALERT pin
*/
/**************************************************************************/
void INA260::setAlertLatch(INA260_AlertLatch state) 
{
    Mask_EnableRegister &= ~(1 << 0);
    Mask_EnableRegister |= ((uint16_t)state << 0);
    i2c_dev->write(INA260_REG_MASK_ENABLE, Mask_EnableRegister, true, true);
}
/**************************************************************************/
/*!
    @brief Checks if the Alert Flag is set
    @return true if the flag is set
*/
/**************************************************************************/
bool INA260::alertFunctionFlag(void) 
{
    i2c_dev->read(INA260_REG_MASK_ENABLE, &Mask_EnableRegister, true, true);
    return (Mask_EnableRegister >> 4) & 1;
}
