
/*
  dircalcs.cpp - geometrics, get inclination, toolface etc from 3D gravitational and/or magnetic sensor readings
  Created by Sicco Dwars, Mar15, 2021.
*/

#include <math.h>
#include <arduino.h>
#include "dircalcs.h"

TriGonioFloats_t TriGonioFloats;

void normalise_to_one_dot_zero_magnitude (Three_D_vector_t *v)
{
    float m;
    m = sqrtf(v->X*v->X + v->Y*v->Y + v->Z*v->Z);
    if (v)
    {
        v->X /= m;
        v->Y /= m;
        v->Z /= m;
    }
}

void populate_TriGonioFloats_from_3Daccel_and_3Dmag (TriGonioFloats_t *TriGonioFloats, Three_D_vector_t *AccelSensors3D, Three_D_vector_t *MagneticSensors3D)
{
    GX = AccelSensors3D->X;
    GY = AccelSensors3D->Y;
    GZ = AccelSensors3D->Z;
    MX = MagneticSensors3D->X;
    MY = MagneticSensors3D->Y;
    MZ = MagneticSensors3D->Z;
   
  
    GX2 = GX * GX;
    GY2 = GY * GY;
    GZ2 = GZ * GZ;

    MX2 = MX * MX;
    MY2 = MY * MY;
    MZ2 = MZ * MZ;

    SQRTGxyz2 = sqrtf(GX2 + GY2 + GZ2);
    SQRTMxyz2 = sqrtf(MX2 + MY2 + MZ2);
}

float RadsToDegs(float x)
{
    return (x * RAD_TO_DEG);
}

float CalculateDeltaAngle (float x_now, float x_previously)
{
    if ((x_now < 90.0) && (x_previously > 270.0))
        return x_now + 360.0 - x_previously;

    if ((x_previously < 90.0) && (x_now > 270.0))
        return x_now - (x_previously+360.0);

    return x_now - x_previously;
}

float KeepAngleWithinPlusOrMinus180 (float x)
{
    if  (x > 180.0)
        return x - 360.0;
    else
        if (x < -180.0)  
             return x + 360.0;
    return x;
}

float mod360 (float x)
{
    if (x > 3600.0)
        x = 3600.0;
    if (x < -3600.0)
        x = -3600.0;
    
    while (x >= 360.0)
        x -= 360.0;

    while (x < 0.0)
        x += 360.0;

    return x;
}

float RADtoDEG_convert__Limit0to360 (float x)
{
    if (isnan(x))
        x = HALF_PI;      // default to 90 deg 
    
    x = RadsToDegs(x);  // convert to degrees 
    return mod360(x);
}

float atan2f_better (float y, float x)
{
    if ((x == 0.0) && (y==0.0))
        return 0.0;
    
    if (fabs(x) >= fabs (y))     // the y/x sum gets noisy if x gets small. Therefore swap x and y and then correct the angle
        return (atan2f (y, x));
    else
    {
        if (y > 0.0)
            return (HALF_PI - (atanf (x / y)));
        else
            return (-HALF_PI - (atanf (x / y)));
    }
}

float Azimuth(TriGonioFloats_t *TriGonioFloats)             
{
    float azimuth;
//    azimuth = atan2f_better ((MY*GX - MX*GY) * SQRTGxyz2, MZ*(GX2 + GY2) - GZ*(GX*MX + GY*MY));
//    azimuth = atan2f_better ((MX*GY - MY*GX) * SQRTGxyz2, MZ*(GX2 + GY2) + GZ*(GX*MX + GY*MY));
    azimuth = atan2f_better ((MY*GX - MX*GY) * SQRTGxyz2, MZ*(GX2 + GY2) + GZ*(GX*MX + GY*MY));
    return RADtoDEG_convert__Limit0to360 (azimuth);                       
}

float DipAngle(TriGonioFloats_t *TriGonioFloats)          
{
    float dradians;
//    dradians = atanf ((GX*MX + GY*MY + GZ*MZ) / (sqrtf( powf((GY*MZ - GZ*MY),2 ) + powf((GX*MZ - GZ*MX),2) + powf((GX*MY - GY*MX),2))));
//    dradians = asinf ((GX*MX + GY*MY + GZ*MZ) / (sqrtf( powf((GY*MZ - GZ*MY),2 ) + powf((GX*MZ - GZ*MX),2) + powf((GX*MY - GY*MX),2))));
    dradians = asinf (GX*MX + GY*MY + GZ*MZ);
    return RADtoDEG_convert__Limit0to360 (dradians);
}

float GravToolFace(TriGonioFloats_t *TriGonioFloats)      
{
    float gravtoolface;
    gravtoolface = atan2f_better( -GX, -GY);  
    return RADtoDEG_convert__Limit0to360 (gravtoolface);
}

float Inclination(TriGonioFloats_t *TriGonioFloats)       
{
    float inclination;
    inclination = atan2f_better(sqrtf( GX2 + GY2 ),GZ);
    return RadsToDegs(inclination); 
}

float MagneticTF(TriGonioFloats_t *TriGonioFloats)                
{
    float magtoolface;
    magtoolface = atan2f_better (-MX,-MY);  
    return RADtoDEG_convert__Limit0to360 (magtoolface);
}

float TotGravFld(TriGonioFloats_t *TriGonioFloats)        
{
    return SQRTGxyz2;     
}

float TotMagFld(TriGonioFloats_t *TriGonioFloats)         
{
    return SQRTMxyz2;
}

int linear_exclusive_scaling(float data, float scale_factor, float low_limit,float high_limit)
{
    int result;
    result = (scale_factor * ((data - low_limit) / (high_limit - low_limit))) + 0.5;
  
    if (result < 0)
        return 0;
    else 
        if (result >= scale_factor-1)
            return scale_factor-1;
    
    return result;
}
