// Copyright © 2019 Richard Gemmell
// Released under the MIT License. See license.txt. (https://opensource.org/licenses/MIT)

#ifndef I2C_DEVICE_H
#define I2C_DEVICE_H

//#include <elapsedMillis.h>
#include "i2c_driver.h"
#ifdef __IMXRT1062__
#include "imx_rt1060_i2c_driver.h"
#endif

// Wraps I2CMaster to make it easy to configure a typical I2C device and then read
// "registers". All calls block so this API should not be called within interrupt
// service routines.

class II2CDevice {
public:
    virtual bool write(uint8_t value, bool send_stop, bool wait_till_done) = 0;
    virtual bool write(uint8_t *buffer, size_t num_bytes, bool send_stop, bool wait_till_done) = 0;

    virtual bool write(uint8_t reg, uint8_t* buffer, size_t num_bytes, bool send_stop, uint32_t swap_bytes_mode, bool wait_till_done) = 0;

    virtual bool write(uint8_t reg, uint8_t value, bool send_stop, bool wait_till_done) = 0;
    virtual bool write(uint8_t reg, int8_t value, bool send_stop, bool wait_till_done) = 0;
    virtual bool write(uint8_t reg, uint16_t value, bool send_stop, bool wait_till_done) = 0;
    virtual bool write(uint8_t reg, int16_t value, bool send_stop, bool wait_till_done) = 0;
    virtual bool write(uint8_t reg, uint32_t value, bool send_stop, bool wait_till_done) = 0;
    virtual bool write(uint8_t reg, int32_t value, bool send_stop, bool wait_till_done) = 0;

    virtual bool read(uint8_t* buffer, size_t num_bytes, bool send_stop, bool wait_till_done) = 0;
    
    virtual bool read(uint8_t reg, uint8_t* buffer, size_t num_bytes, bool send_stop, uint32_t swap_bytes_mode, bool wait_till_done) = 0;
    
    virtual bool read(uint8_t reg, uint8_t* value, bool send_stop, bool wait_till_done) = 0;
    virtual bool read(uint8_t reg, int8_t* value, bool send_stop, bool wait_till_done) = 0;
    virtual bool read(uint8_t reg, uint16_t* value, bool send_stop, bool wait_till_done) = 0;
    virtual bool read(uint8_t reg, int16_t* value, bool send_stop, bool wait_till_done) = 0;
    virtual bool read(uint8_t reg, uint32_t* value, bool send_stop, bool wait_till_done) = 0;
    virtual bool read(uint8_t reg, int32_t* value, bool send_stop, bool wait_till_done) = 0;

/*    
    virtual uint16_t swap_if_needed16 (uint16_t x) = 0;
    virtual uint32_t swap_if_needed32 (uint32_t x) = 0;
    virtual int16_t swap_if_needed16i (int16_t x) = 0;
    virtual int32_t swap_if_needed32i (int32_t x) = 0;
*/
    
};

// Wraps I2CMaster to make it easy to configure a typical I2C device and then read
// "registers". All calls block so this API should not be called within interrupt
// service routines.

class I2CDevice : public II2CDevice {
public:
    const uint32_t timeout_millis = 200;

    I2CDevice(IMX_RT1060_I2CMaster& master, uint8_t address, int device_byte_order = _LITTLE_ENDIAN)
            : master(master), address(address) {
        swap_bytes = _BYTE_ORDER != device_byte_order;
    }

    bool write(uint8_t reg, uint8_t* buffer, size_t num_bytes, bool send_stop, uint32_t swap_bytes_mode, bool wait_till_done) override {
        uint8_t big_buffer[num_bytes+1];
        big_buffer[0] = reg;
        memcpy(big_buffer+1, buffer, num_bytes);
        if (!swap_bytes) 
            swap_bytes_mode = 0;
         
        I2C_transaction_t* p = master.write_async_q(address, big_buffer, sizeof(big_buffer), swap_bytes_mode, send_stop);
        if (wait_till_done)
            finish(p);
        return !master.has_error();
    }

    inline bool write(uint8_t value, bool send_stop, bool wait_till_done) override
    {
        uint8_t big_buffer[1];
        big_buffer[0] = value;
        I2C_transaction_t* p = master.write_async_q(address, big_buffer, 1, 0, send_stop);
        if (wait_till_done)
            finish(p);
        return !master.has_error();
    }

    inline bool write(uint8_t *buffer, size_t num_bytes, bool send_stop, bool wait_till_done) override
    {
        uint8_t big_buffer[num_bytes];
        memcpy(big_buffer, buffer, num_bytes);
        I2C_transaction_t* p = master.write_async_q(address, big_buffer, num_bytes, 0, send_stop);
        if (wait_till_done)
            finish(p);
        return !master.has_error();
    }
    
    inline bool write(uint8_t reg, uint8_t value, bool send_stop, bool wait_till_done) override {
        return write(reg, &value, 1, send_stop, 0, wait_till_done);
    }

    inline bool write(uint8_t reg, int8_t value, bool send_stop, bool wait_till_done) override {
        return write(reg, (uint8_t*)&value, 1, send_stop, 0, wait_till_done);
    }

    inline bool write(uint8_t reg, uint16_t value, bool send_stop, bool wait_till_done) override {
            return write(reg, (uint8_t*)&value, 2, send_stop, 1, wait_till_done);
    }

    inline bool write(uint8_t reg, int16_t value, bool send_stop, bool wait_till_done) override {
            return write(reg, (uint8_t*)&value, 2, send_stop, 1, wait_till_done);
    }

    inline bool write(uint8_t reg, uint32_t value, bool send_stop, bool wait_till_done) override {
            return write(reg, (uint8_t*)&value, 4, send_stop, 2, wait_till_done);
    }

    inline bool write(uint8_t reg, int32_t value, bool send_stop, bool wait_till_done) override {
            return write(reg, (uint8_t*)&value, 4, send_stop, 2, wait_till_done);
    }

    bool read(uint8_t reg, uint8_t* buffer, size_t num_bytes, bool send_stop, uint32_t swap_bytes_mode, bool wait_till_done) override {
        bool has_error = true;
        if (write_register(reg)) 
        {
            if (!swap_bytes) 
                swap_bytes_mode = 0;

            I2C_transaction_t* p = master.read_async_q(address, buffer, num_bytes, swap_bytes_mode, send_stop);
            if (wait_till_done)
                finish(p);
            has_error = master.has_error();
        }
        if (has_error) {
            // Zero the buffer if the read failed to avoid using stale data.
            memset(buffer, 0, num_bytes);
        }
        return !has_error;
    }
    
    inline bool read(uint8_t* buffer, size_t num_bytes, bool send_stop, bool wait_till_done) override
    {
        bool has_error = true;
        I2C_transaction_t* p = master.read_async_q(address, buffer, num_bytes, 0, send_stop);
        if (wait_till_done)
            finish(p);
        has_error = master.has_error();
        if (has_error) {
            // Zero the buffer if the read failed to avoid using stale data.
            memset(buffer, 0, num_bytes);
        }
        return !has_error;
    }

    inline bool read(uint8_t reg, uint8_t* value, bool send_stop, bool wait_till_done) override {
        return read(reg, value, 1, send_stop, 0, wait_till_done);
    }

    inline bool read(uint8_t reg, int8_t* value, bool send_stop, bool wait_till_done) override {
        return read(reg, (uint8_t*)value, 1, send_stop, 0, wait_till_done);
    }

    inline bool read(uint8_t reg, uint16_t* value, bool send_stop, bool wait_till_done) override {
        return read(reg, (uint8_t*)value, 2, send_stop, 1, wait_till_done);
    }

    inline bool read(uint8_t reg, int16_t* value, bool send_stop, bool wait_till_done) override {
        return read(reg, (uint8_t*)value, 2, send_stop, 1, wait_till_done);
    }

    inline bool read(uint8_t reg, uint32_t* value, bool send_stop, bool wait_till_done) override {
        return read(reg, (uint8_t*)value, 4, send_stop, 2, wait_till_done);
    }

    inline bool read(uint8_t reg, int32_t* value, bool send_stop, bool wait_till_done) override {
        return read(reg, (uint8_t*)value, 4, send_stop, 2, wait_till_done);
    }

private:
    IMX_RT1060_I2CMaster& master;
    uint8_t address;
    bool swap_bytes;

    bool write_register(uint8_t reg) {
        I2C_transaction_t* p = master.write_async_q(address, &reg, 1, 0, false);
        finish(p);
        return !master.has_error();
    }

    void finish(I2C_transaction_t *p) {
        elapsedMillis timeout;
        while (timeout <= timeout_millis) {
            if (p->completed) {
                return;
            }
        }
    }
};

#endif //I2C_DEVICE_H
